import sys

from termcolor import colored, cprint


def term_columns():
    # columns, lines = get_terminal_size()  # pylint: disable=W0612
    columns = 80
    return columns


def print_error(message):
    return cprint(message, "red", attrs=["bold"], file=sys.stderr)


def print_info(*args):
    assert args
    if len(args) > 1:
        print(args[0], info(" ".join((args[1:]))))
    else:
        cprint(args[0], "cyan")


def print_success(message):
    return cprint(message, "green", attrs=["bold"])


def print_header(message):
    return cprint(message, "white", attrs=["bold"])


def print_warn(message):
    return cprint(message, "yellow", attrs=["bold"], file=sys.stderr)


def info(message):
    return colored(message, "cyan")


def info_header(message):
    return colored(message, "cyan", attrs=["bold"])


def warning(message):
    return colored(message, "yellow")


def success(message):
    return colored(message, "green", attrs=["bold"])


def error(message):
    return colored(message, "red", attrs=["bold"])


def print_list(label, items):
    if not items:
        return
    print_header(label)
    if isinstance(items, (list, tuple)):
        for item in items:
            print_info("\t" + item)
    if isinstance(items, dict):
        for key, value in items.items():
            if value:
                if isinstance(value, str):
                    print("{0:>15}: {1}".format(key, info(value)))
                if isinstance(value, list):
                    print("{0:>15}:".format(key))
                    for value_item in value:
                        print(" " * 15, info(value_item))
                if isinstance(value, dict):
                    print("{0:>15}:".format(key))
                    for dict_key, dict_value in value.items():
                        print(
                            " " * 16
                            + "{0:>15}: {1}".format(
                                info_header(dict_key), info(dict_value)
                            )
                        )



""" Provide information about container images available from LXC """
from typing import Optional

import typer

from ..colorhelper import print_list
from ..lxc import LCXMetaData

app = typer.Typer()


@app.command()
def list(item: Optional[str] = typer.Argument(None)):
    lcx_meta = LCXMetaData()
    print_list("Distros", lcx_meta.distros())


@app.command()
def info(item: str):
    print(f"info item: {item}")



import os

from prenv.mount import MS_BIND, MS_PRIVATE, MS_REC, mount
from prenv.unshare import set_user_level_root

set_user_level_root()

# remount / as private to make subsequent mounts visible only to the current process and its descendants.
mount(None, "/", None, MS_REC | MS_PRIVATE)

os.chdir("/tmp/alpine")
mount("/proc", "proc/", None, MS_BIND | MS_REC)
mount("/dev", "dev/", None, MS_BIND | MS_REC)
mount("/sys", "sys/", None, MS_BIND | MS_REC)
mount("tmpfs", "run/", "tmpfs")

os.chroot(".")

# Spawn a shell
os.execv("/bin/sh", ["/bin/sh"])



import requests

from .verbose import verbose

LCX_INDEX = "https://images.linuxcontainers.org/meta/1.0/index-user"
LXC_URL_TEMPL = "https://images.linuxcontainers.org/images/{}/{}/{}/{}/{}/rootfs.tar.xz"


class NotSingleVersonError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class LCXMetaData:
    def __init__(self):
        verbose(f"Fetching LXC metadata from {LCX_INDEX}")
        reply = requests.get(LCX_INDEX)
        reply.raise_for_status()
        verbose(f"Received {reply.status_code} {reply.reason} from {reply.url}")
        self._index = self.csv_to_dict(reply.text)

    @staticmethod
    def csv_to_dict(csv):
        lines = csv.splitlines()
        header = ("name", "version", "arch", "variant", "build", "path")
        return [dict(zip(header, line.split(";"))) for line in lines]

    def distros(self):
        return tuple(set([item["name"] for item in self._index]))

    def versions(self, distro_name, distro_version, distro_arch, distro_variant):
        return tuple(
            set(
                [
                    item["version"]
                    for item in self._index
                    if item["name"] == distro_name
                    and item["arch"] == distro_arch
                    and item["variant"] == distro_variant
                    and (item["version"] == distro_version if distro_version else True)
                ]
            )
        )

    def builds(self, distro_name, distro_version, distro_arch, distro_variant):
        return tuple(
            set(
                [
                    item["build"]
                    for item in self._index
                    if item["name"] == distro_name
                    and item["arch"] == distro_arch
                    and item["variant"] == distro_variant
                    and item["version"] == distro_version
                ]
            )
        )

    def image_url(
        self,
        distro_name,
        distro_version=None,
        distro_arch="amd64",
        distro_variant="default",
    ):
        """Return the URL for the given distro"""
        matching_versions = self.versions(
            distro_name, distro_version, distro_arch, distro_variant
        )
        if len(matching_versions) == 0:
            raise ValueError(
                f"No image found matching {distro_name} {distro_version} {distro_arch} {distro_variant}"
            )
        if len(matching_versions) > 1:
            raise NotSingleVersonError(
                f"Found multiple versions for {distro_name} {distro_version} {distro_arch}",
                matching_versions,
            )
        matching_version = matching_versions[0]
        matchin_builds = self.builds(
            distro_name, matching_version, distro_arch, distro_variant
        )
        assert len(matchin_builds) == 1
        matching_build = matchin_builds[0]
        url = LXC_URL_TEMPL.format(
            distro_name, matching_version, distro_arch, distro_variant, matching_build
        )
        verbose(f"Found image URL {url}")
        return url



import ctypes
import ctypes.util
import os
from pathlib import Path
from typing import Union

libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)
libc.mount.argtypes = (
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_char_p,
    ctypes.c_ulong,
    ctypes.c_char_p,
)
libc.mount.restype = ctypes.c_int
libc.umount.argtypes = (ctypes.c_char_p,)
libc.umount.restype = ctypes.c_int

MS_RDONLY = 0x1
MS_NOSUID = 0x2
MS_NODEV = 0x4
MS_NOEXEC = 0x8
MS_SYNCHRONOUS = 0x10
MS_REMOUNT = 0x20
MS_MANDLOCK = 0x40
MS_DIRSYNC = 0x80
MS_NOATIME = 0x400
MS_NODIRATIME = 0x800
MS_BIND = 0x1000
MS_MOVE = 0x2000
MS_REC = 0x4000
MS_SILENT = 0x8000
MS_UNBINDABLE = 0x20000
MS_PRIVATE = 0x40000
MS_SLAVE = 0x80000
MS_SHARED = 0x100000
MS_RELATIME = 0x200000
MS_STRICTATIME = 0x1000000
MS_LAZYTIME = 0x2000000


def mount(
    device: Union[str, Path],
    mount_point: Union[str, Path],
    fs_type: str,
    flags: int = 0,
    options: str = "",
):
    ret = libc.mount(
        str(device).encode(),
        str(mount_point).encode(),
        fs_type.encode() if fs_type else None,
        flags,
        options.encode(),
    )
    if ret == -1:
        errno = ctypes.get_errno()
        raise OSError(
            errno,
            f'Fail to mount {device} ({fs_type}) on {mount_point} with "{options}": {os.strerror(errno)}',
        )


def umount(target: Union[str, Path]):
    ret = libc.umount(str(target).encode())
    if ret == -1:
        errno = ctypes.get_errno()
        raise OSError(errno, f"Fail to umount {target}: {os.strerror(errno)}")



from tempfile import NamedTemporaryFile

from drclient.cli.cmd_pull import pull
from drclient.registry import DockerRegistryClient

from .download import Cache


def pull_docker_image(image_name):
    registry, repository, tag = DockerRegistryClient.parse_image_url(image_name)
    source_reference = f"{registry}/{repository}:{tag}"
    return download(source_reference)


def download(image_url, use_cache=True):
    # Check if image is on cache
    if use_cache:
        cache = Cache()
        cached_image = cache.get(image_url)
        if cached_image:
            cache_fn, last_modified, file_size = cached_image
            return cache_fn
    with NamedTemporaryFile(delete=False) as tmp_file:
        pull(image_url, tar_file=tmp_file.name, output_directory=None)
        tmp_file.flush()
    if use_cache:
        return cache.put(tmp_file.name, image_url)
    return tmp_file.name



import atexit
import os
import socket
import time
from pathlib import Path

import xdg

SOCKETS_DIR = Path(xdg.xdg_runtime_dir(), "prenv")


def create_socket_bind():
    my_pid = os.getpid()
    SOCKETS_DIR.mkdir(parents=True, exist_ok=True)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    socket_path = Path(SOCKETS_DIR, f"prenv.{my_pid}.sock")
    atexit.register(os.unlink, socket_path.as_posix())
    sock.bind(socket_path.as_posix())
    return sock


def get_process_socket(pid: int) -> socket.socket:
    socket_path = Path(SOCKETS_DIR, f"prenv.{pid}.sock")

    # Wait for the child process to create the Unix socket
    while not socket_path.exists():
        time.sleep(0.01)

    conn = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    conn.connect(socket_path.as_posix())
    return conn



from __future__ import print_function

import tarfile


def extract_tar(tar_fname, dest_dir):
    tarfile.os.chown = (
        lambda x, y, z: 0
    )  # Monkey patch chown because we don't care about ownership

    tarfile.os.mknod = (
        lambda x, y, z: 0
    )  # Monkey patch mknod because some layers include devices
    with tarfile.open(tar_fname) as image_tar:
        image_tar.extractall(dest_dir, numeric_owner=True)



import hashlib
import os
import shutil
from datetime import datetime
from os.path import basename, exists, expanduser, join
from tempfile import NamedTemporaryFile

import requests
from dateutil.parser import parse as parsedate
from rich.progress import Progress

from .colorhelper import print_error, print_info, print_warn
from .size import HumanSize
from .verbose import verbose

CACHE_PATH = join(expanduser("~"), ".prenv", "images_cache")


def url_to_filename(url: str):
    """Convert url to filename"""
    url = url.replace("://", "_")
    url = url.replace("/", "_")
    url = url.replace(":", "_")
    return url


class Cache(object):
    cache_dir = CACHE_PATH
    """ Provides an image caching mechanism on disk """

    def __init__(self):
        if not exists(CACHE_PATH):
            os.makedirs(CACHE_PATH, 0o700)

    def get(self, cache_key, default=None):
        """return info for cached file"""

        cache_fn = join(CACHE_PATH, url_to_filename(cache_key))

        if exists(cache_fn):
            file_stat = os.stat(cache_fn)
            last_modified = datetime.fromtimestamp(file_stat.st_mtime)
            file_size = file_stat.st_size
            return cache_fn, last_modified, file_size

        return default

    def put(self, filename, cache_key):
        """put a file into cache"""
        cache_fn = join(CACHE_PATH, url_to_filename(cache_key))
        shutil.move(filename, cache_fn)
        return cache_fn


def download(image_url, use_cache=True):
    """Download image (if not found in cache) and return it's filename"""

    response = requests.head(image_url, allow_redirects=True)
    response.raise_for_status()
    file_size = remote_file_size = int(response.headers.get("Content-Length"))
    remote_last_modified = parsedate(response.headers.get("Last-Modified")).replace(
        tzinfo=None
    )
    remote_is_valid = response.status_code == 200 and file_size and remote_last_modified

    # Check if image is on cache
    if use_cache:
        cache = Cache()
        cached_image = cache.get(image_url)
    else:
        cached_image = None
    if cached_image:
        if remote_is_valid:
            cache_fn, last_modified, file_size = cached_image
            if remote_file_size == file_size and remote_last_modified < last_modified:
                verbose("Using file from cache", CACHE_PATH)
                return cache_fn
            verbose("Checking new remote file because an update was found")
        else:
            print_warn("Unable to check the status for " + image_url)
            print_warn("Assuming local cache is valid")

    # Not cached, and no valid remote information was found
    if not remote_is_valid:
        print_error(
            "Unable to get file, http_code=%s, size=%s, last_modified=%s"
            % (response.status_code, remote_file_size, remote_last_modified)
        )
        exit(2)

    # Dowload image
    print_info(
        "Downloading image... ",
        "{0} [{1:.2S}]".format(basename(image_url), HumanSize(remote_file_size)),
    )
    remote_sha256 = hashlib.sha256()
    response = requests.get(image_url, stream=True)
    response.raise_for_status()
    with NamedTemporaryFile(delete=False) as tmp_file:
        with Progress() as progress:
            task1 = progress.add_task("[green]Processing...", total=remote_file_size)
            for chunk in response.iter_content(chunk_size=1024):
                remote_sha256.update(chunk)
                tmp_file.write(chunk)
                progress.update(task1, advance=len(chunk))
            tmp_file.flush()
    if use_cache:
        return cache.put(tmp_file.name, image_url)
    else:
        return tmp_file.name



import typer

from .cli import cmd_create, cmd_exec, cmd_lxc, cmd_run, verbose


def main():
    app = typer.Typer(short_help="Utility to manage a pypy.org metadata cache")

    app.command(help="Create a container instance")(cmd_create.create)
    app.command(help="Execute a command in an existing instance")(cmd_exec.exec)
    app.command(help="Run a command in an ephemeral instance")(cmd_run.run)
    app.add_typer(
        cmd_lxc.app, name="lxc", help="Get information about available LCX images"
    )

    app.callback()(verbose.verbose)
    app()


if __name__ == "__main__":
    main()



import typer

is_verbose = False


def verbose(
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Enable verbose mode")
):
    if verbose:
        print("Running in verbose mode")
        global is_verbose
        is_verbose = True



import os
from pathlib import Path
from tempfile import mkdtemp

from .mount import MS_BIND, MS_PRIVATE, MS_REC, mount
from .tar import extract_tar
from .unshare import set_user_level_root

STD_MOUNTS = [
    ("/proc", "proc/", None, MS_BIND | MS_REC),
    ("/dev", "dev/", None, MS_BIND | MS_REC),
    ("/sys", "sys/", None, MS_BIND | MS_REC),
    ("tmpfs", "run/", "tmpfs"),
]


def prepare_rootfs(rootfs_filename: str, in_memory):
    set_user_level_root()
    mount(None, "/", None, MS_REC | MS_PRIVATE)
    mount_dir = Path(mkdtemp())
    if in_memory:
        mount("tmpfs", mount_dir, "tmpfs")
    extract_tar(rootfs_filename, mount_dir)
    os.chdir(mount_dir)
    for mount_args in STD_MOUNTS:
        os.makedirs(mount_args[1], exist_ok=True)
        mount(*mount_args)
    resolv_conf = Path(f"{mount_dir}/etc/resolv.conf")
    if resolv_conf.is_symlink():
        resolv_conf.unlink()
    if Path(mount_dir, "etc").exists():
        resolv_conf.touch()
        mount("/etc/resolv.conf", f"{mount_dir}/etc/resolv.conf", None, MS_BIND)
    return mount_dir



import ctypes
import os
import time
from pathlib import Path

from .verbose import verbose

CLONE_NEWNS = 0x20000
CLONE_NEWCGROUP = 0x2000000
CLONE_NEWUTS = 0x4000000
CLONE_NEWIPC = 0x8000000
CLONE_NEWUSER = 0x10000000
CLONE_NEWPID = 0x20000000
CLONE_NEWNET = 0x40000000
CLONE_NEWTIME = 0x80000000


libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)

libc.unshare.argtypes = (ctypes.c_int,)
libc.unshare.restype = ctypes.c_int


def unshare(flags: int):
    ret = libc.unshare(flags)
    if ret == -1:
        errno = ctypes.get_errno()
        raise OSError(
            errno,
            f"Fail to unshare with flags {flags:#x}: {os.strerror(errno)}",
        )


def rewrite_uid_map(uid, gid):
    Path("/proc/self/uid_map").write_text(f"0 {uid} 1")
    Path("/proc/self/setgroups").write_text("deny")
    Path("/proc/self/gid_map").write_text(f"0 {gid} 1")
    # wait for the uid/gid mapping to take effect
    time.sleep(0.1)


def set_user_level_root():
    uid, gid = os.geteuid(), os.getegid()
    verbose("Creating user and mount namespaces")
    unshare(CLONE_NEWNS | CLONE_NEWUSER)
    rewrite_uid_map(uid, gid)



"""  Create a new environment  """
import os
from typing import Optional

import typer

from ..docker import pull_docker_image
from ..download import download
from ..rootfs import prepare_rootfs
from ..verbose import verbose
from .cmd_create import get_distro_url


def run(
    distro_name: str = typer.Argument(...),
    no_shell: bool = typer.Option(False, "--no-sh"),
    from_docker: bool = typer.Option(
        False, "--from-docker", help="Run from docker image"
    ),
    command: Optional[str] = typer.Argument("/bin/sh", help="Command to be run"),
    chroot_break: bool = typer.Option(False, "--break", help="Break into shell"),
):
    if from_docker:
        tar_fname = pull_docker_image(distro_name)
    else:
        distro_url = get_distro_url(distro_name)
        verbose(f"Checking file {distro_url} for {distro_name}")
        tar_fname = download(distro_url)

    mount_dir = prepare_rootfs(tar_fname, in_memory=True)
    if chroot_break:
        os.chdir(mount_dir)
        os.system("bash")
        exit(0)
    os.chroot(mount_dir)
    os.chdir("/")
    if no_shell:
        command = command.split()
        os.execvp(command[0], command[:])
    else:
        os.execvp("/bin/sh", ["sh", "-c", command])



from __future__ import print_function

import math
import sys


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class HumanSize(int):
    """define a size class to allow custom formatting
    Implements a format specifier of S for the size class - which displays a human readable in b, kb, Mb etc
    """

    def __format__(self, fmt):
        if fmt == "" or fmt[-1] != "S":
            if fmt[-1].tolower() in ["b", "c", "d", "o", "x", "n", "e", "f", "g", "%"]:
                # Numeric format.
                return int(self).__format__(fmt)
            return str(self).__format__(fmt)

        val, text = float(self), ["b ", "Kb", "Mb", "Gb", "Tb", "Pb"]
        if val < 1:
            # Can't take log(0) in any base.
            i, j = 0, 0
        else:
            i = int(math.log(val, 1024)) + 1
            j = val / math.pow(1024, i)
            j, i = (j, i) if j > 0.5 else (j * 1024, i - 1)
        return ("{0:{1}f}" + text[i]).format(j, fmt[:-1])



"""  Create a new environment  """
import os
import sys
from typing import Optional

import typer

from ..docker import pull_docker_image
from ..download import download
from ..lxc import LCXMetaData
from ..process import create_child_process
from ..verbose import verbose


def get_distro_url(distro_name: str):
    lcx = LCXMetaData()
    distro_version = None
    if ":" in distro_name:
        distro_name, distro_version = distro_name.split(":")
    if distro_name not in lcx.distros():
        raise typer.BadParameter(f"Unknown distro {distro_name}")
    return lcx.image_url(distro_name, distro_version=distro_version)


def create(
    distro_name: str = typer.Argument(...),
    no_shell: bool = typer.Option(False, "--no-sh"),
    from_docker: bool = typer.Option(
        False, "--from-docker", help="Run from docker image"
    ),
    command: Optional[str] = typer.Argument(None, help="Command to be run"),
):
    if from_docker:
        tar_fname = pull_docker_image(distro_name)
    else:
        distro_url = get_distro_url(distro_name)
        verbose(f"Checking file {distro_url} for {distro_name}")
        tar_fname = download(distro_url)
    pid = create_child_process(tar_fname)
    if command:
        args = [sys.executable, "-m", "prenv", "join", str(pid), command]
        # Respect the no-sh behavior
        if no_shell:
            args.insert(4, "--no-sh")
        os.execvp(sys.executable, args)



import os

from .colorhelper import info, print_error
from .rootfs import prepare_rootfs
from .socket import create_socket_bind, get_process_socket
from .verbose import verbose


def child_process(tar_fname) -> None:
    sock = create_socket_bind()
    sock.listen(1)
    conn, _ = sock.accept()

    message = conn.recv(1024).decode("utf-8")
    if message != "setup":
        raise Exception(f"Unexpected message from parent process: {message}")
    verbose(f"Received message from parent process: {message}")
    mount_dir = prepare_rootfs(tar_fname, in_memory=True)
    conn.sendall(b"ok")
    conn.close()

    while True:
        conn, _ = sock.accept()
        message = conn.recv(1024).decode("utf-8")
        if message == "info":
            conn.sendall(mount_dir.as_posix().encode("utf-8"))
        else:
            print("Received unexpeted message from parent process: ", message)
        conn.close()


def parent_process(pid: int) -> None:
    # Connect to the child process's Unix socket
    conn = get_process_socket(pid)
    conn.sendall(b"setup")

    # Wait for the setup result
    message = conn.recv(1024).decode("utf-8")
    if message == "ok":
        print("Instance setup successfully with PID:", info(pid))
    else:
        print_error(f"Failed to setup the instance: {message}")
        exit(2)


def create_child_process(tar_fname: str) -> None:
    """Create a master process that will create a new process namespace and
    mount a new root filesystem."""
    pid = os.fork()

    if pid == 0:
        child_process(tar_fname)
    else:
        parent_process(pid)
        return pid



import ctypes
import os

from .socket import get_process_socket
from .unshare import CLONE_NEWNS, CLONE_NEWUSER

SYSCALL_PIDFD_OPEN = 434

libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)


def enter_process(pid: int, no_shell, command) -> None:
    """Enter the process namespace of a given PID."""

    fd = libc.syscall(SYSCALL_PIDFD_OPEN, pid, 0, None, None, None)
    if fd < 0:
        print("pidfd_open failed: %s", ctypes.get_errno())
        return

    if libc.setns(fd, CLONE_NEWUSER | CLONE_NEWNS) < 0:
        errno = ctypes.get_errno()
        raise OSError(
            errno,
            f"setns failed: {os.strerror(errno)}",
        )

    conn = get_process_socket(pid)
    conn.sendall(b"info")
    data = conn.recv(1024)
    conn.close()
    os.chroot(data.decode("utf-8"))
    os.chdir("/")
    if no_shell:
        command = command.split()
        os.execvp(command[0], command[:])
    else:
        os.execvp("/bin/sh", ["sh", "-c", command])



"""  Create a new environment  """

from typing import Optional

import typer

from ..enter import enter_process


def exec(
    pid: int,
    no_shell: bool = typer.Option(False, "--no-sh"),
    command: Optional[str] = typer.Argument("/bin/sh", help="Command to be run"),
):
    enter_process(pid, no_shell, command)
